#pragma once

#include<dune/pdelab/common/quadraturerules.hh>
#include<dune/pdelab/finiteelement/localbasiscache.hh>
//#include<dune/pdelab/localoperator/idefault.hh>
#include<dune/pdelab/localoperator/defaultimp.hh>
#include<dune/pdelab/localoperator/pattern.hh>
#include<dune/pdelab/localoperator/flags.hh>
#include<dune/pdelab/localoperator/variablefactories.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>

#include<dune/phasefield/localoperator/fem_base_chns.hh>



template<typename Param, typename CHGFS, typename CHContainer, typename NSGFS, typename NSContainer>
    class FEM_FS_CHNS_R_ThinPlateFlow :
      public Dune::PDELab::FullVolumePattern,
      public Dune::PDELab::LocalOperatorDefaultFlags,
      public Dune::PDELab::NumericalJacobianVolume<FEM_FS_CHNS_R_ThinPlateFlow<Param, CHGFS, CHContainer, NSGFS, NSContainer> >,
      public Dune::PDELab::NumericalJacobianApplyVolume<FEM_FS_CHNS_R_ThinPlateFlow<Param, CHGFS, CHContainer, NSGFS, NSContainer> >
    {
    private:
      typedef typename CHContainer::field_type RF;
      typedef typename NSGFS::Traits::GridViewType::template Codim<0>::Entity   ENTITY;

      CHLocalBasisCache<CHGFS> chCache;
      VLocalBasisCache<NSGFS> vCache;
      PLocalBasisCache<NSGFS> pCache;

      typedef GFS_LocalViewer<CHGFS, CHContainer, ENTITY, RF> CHLView;
      mutable CHLView chLocal;
      //mutable CHLView chOldLocal;

      typedef GFS_LocalViewer<NSGFS, NSContainer, ENTITY, RF> NSLView;
      mutable NSLView nsOldLocal;

      Param& param; // parameter functions


    public:


      // pattern assembly flags
      enum { doPatternVolume = true };

      // residual assembly flags
      enum { doAlphaVolume = true };


      FEM_FS_CHNS_R_ThinPlateFlow (Param& param_, CHGFS& chGfs_, CHContainer& chV_,
                         NSGFS& nsGfs_, NSContainer& nsVOld_) :
        chLocal(chGfs_,chV_),
        //chOldLocal(chGfs_,chVOld_),
        nsOldLocal(nsGfs_,nsVOld_),
        param(param_)
      {
      }


      template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
      void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
      {
        //Quadrature Rule
        const int dim = EG::Entity::dimension;
        const int order = 5;
        auto rule = Dune::PDELab::quadratureRule(eg.geometry(), order);

        chLocal.DoEvaluation(eg);
        //chOldLocal.DoEvaluation(eg);
        nsOldLocal.DoSubEvaluation(eg,child(nsOldLocal.lfs,Dune::Indices::_0));

        for(const auto& ip : rule)
        {
          const auto S = eg.geometry().jacobianInverseTransposed(ip.position());
          auto factor = ip.weight()*eg.geometry().integrationElement(ip.position());

          CHVariables<RF,dim,decltype(chLocal.lfs)> ch(chCache, ip.position(), chLocal.lfs, chLocal.lv, S);
          NSVariables_P_VOld_Extension<RF,dim,LFSV> ns(vCache, pCache, ip.position(), lfsv, x.base(), nsOldLocal.lv, S);

          RF rho_f=param.phys.rho_1*ch.pf;
          RF rho_f_reg= rho_f + param.delta;

          RF pf_f = ch.pf + 2*param.delta*(1-ch.pf);
          Dune::FieldVector<RF,dim> gradpf_f(0.0);
          gradpf_f.axpy((1-2*param.delta),ch.gradpf);

          RF mu_reg = pf_f*param.phys.mu;

          for (size_t i=0; i<ns.pspace.size(); i++)
          {
            //r.accumulate(pspace,i, -1.0* pf_f*(v*gradphi[i][0])*factor);
            r.accumulate(ns.pspace,i, -1.0 * (pf_f*ns.div_v  + (gradpf_f*ns.v) )* ns.pbasis.phi[i] * factor);
          }

          for (size_t i=0; i<ns.vfemspace.size(); i++)
          {
            for(int k=0; k<dim; k++)
            {
              r.accumulate(child(ns.vspace,k),i, (   rho_f_reg*(ns.v[k]-ns.vOld[k])*ns.vbasis.phi[i]/param.time.dt          //TimeEvolution
                                                //- ns.p * (pf_f*ns.vbasis.gradphi[i][0][k] + ns.vbasis.phi[i]*gradpf_f[k])  //Pressure in weak form
                                                + pf_f * (ns.gradp[k] * ns.vbasis.phi[i])                           //Pressure in strong form
                                                + mu_reg * (ns.jacv[k]*ns.vbasis.gradphi[i][0])        //Viscosity
                                                + param.vDis.eval(pf_f)*ns.v[k]*ns.vbasis.phi[i]              //Velocity Dissipation in Solid
                                                //- rho*g[k]*vphi[i]                              //Gravitation
                                                + mu_reg * 8 / (param.CellHeight * param.CellHeight) * ns.v[k]*ns.vbasis.phi[i] //Drag from bottom and top plate
                                              )*factor);
            // integrate (grad u)^T * grad phi_i                                                  //Transpose for Viscosity
            for(int dd=0; dd<dim; ++dd)
                r.accumulate(child(ns.vspace,k),i, param.phys.mu * (ns.jacv[dd][k] * (ns.vbasis.gradphi[i][0][dd])) * factor);
            }
          }
        }
      }

    };
