#pragma once

#include<dune/common/parametertreeparser.hh>

#include<dune/pdelab/finiteelementmap/p0fem.hh>
#include <dune/pdelab/adaptivity/adaptivity.hh>
#include<dune/pdelab/gridoperator/gridoperator.hh>
#include<dune/pdelab/backend/istl.hh>

#include <dune/phasefield/base_parameters.hh>
#include <dune/phasefield/chns_base.hh>

//_________________________________________________________________________________________________________________________

class SummerschoolTimestepManager :
public TimestepManager
{
public:
  double reactionOnlyDt;
  bool lastTimestepReactionOnly;
  SummerschoolTimestepManager(const Dune::ParameterTree & param, Params_Time& time_, int verbosity) :
  TimestepManager(param, time_, verbosity)
  {
    reactionOnlyDt = param.get("reactionOnlyDt",(double)30);
    lastTimestepReactionOnly = false;
  }

  template<typename REACTION_NEWTON, typename VTKWRITER>
  int doReactionOnlyTimestep(REACTION_NEWTON& reaction_newton, VTKWRITER& vtkwriter)
  {
    try
    {
      if (verb > 1)   std::cout << "= Begin newtonapply REACTION:" << std::endl;

      double tmpdt = time.dt;
      time.dt = reactionOnlyDt;
      reaction_newton.apply();
      time.dt = tmpdt;
    }
    catch(...)
    {
      std::cout << "===============================" << std::endl;
      std::cout << "Problem with Newton Solver, aborting :(" << std::endl;
      std::cout << "===============================" << std::endl;
      return -1;
    }

    lastTimestepReactionOnly = true;
    time.t += reactionOnlyDt;

    if( time.t < saveintervallstart || time.t>lastsave+2*saveintervall ) // Write reaction steps only if there are no other steps inbetween
    {
        lastsave = time.t;
        vtkwriter.write(time.t,Dune::VTK::ascii);
    }
    return 1;
  }

};
